from typing import TYPE_CHECKING

from pendulum.datetime import DateTime
from sqlalchemy import ForeignKey, Integer, String
from sqlalchemy.orm import Mapped, mapped_column, relationship

from duty_board.alchemy.settings import Base
from duty_board.alchemy.sqlalchemy_types import UtcDateTime
from duty_board.models.calendar import Calendar

if TYPE_CHECKING:
    from duty_board.models.person import Person


class OnCallEvent(Base):
    """
    SQLAlchemy Model for on-call-events. Meaning, the time a person is on Duty.

    Note: Originally the uid was not added, however, sqladmin does not support composite keys. So this was added to be
    a singular primary key.
    """

    __tablename__ = "on_call_event"
    uid: Mapped[int] = mapped_column(primary_key=True, autoincrement=True, unique=True)
    start_event_utc: Mapped[DateTime] = mapped_column(UtcDateTime(), nullable=False)
    end_event_utc: Mapped[DateTime] = mapped_column(UtcDateTime(), nullable=False)

    calendar_uid: Mapped[str] = mapped_column(String(50), ForeignKey("calendar.uid", ondelete="CASCADE"))
    calendar: Mapped["Calendar"] = relationship(back_populates="events")

    person_uid: Mapped[int] = mapped_column(Integer, ForeignKey("person.uid", ondelete="CASCADE"), nullable=False)
    person: Mapped["Person"] = relationship(back_populates="upcoming_events")

    def __repr__(self) -> str:
        return (
            f"OnCallEvent(calendar='{self.calendar_uid}', start_event='{self.start_event_utc}', "
            f"end_event='{self.end_event_utc}', person='{self.person_uid}')"
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import contextlib
from typing import Generator

from sqlalchemy.orm import Session as SASession

from duty_board.alchemy import settings


@contextlib.contextmanager
def create_session() -> Generator[SASession, None, None]:
    """Contextmanager that will create and teardown a session."""
    if not settings.Session:  # type: ignore[truthy-function]
        raise RuntimeError("Session must be set before!")
    session = settings.Session()
    try:
        yield session
        session.commit()
    except Exception:
        session.rollback()
        raise
    finally:
        session.close()



from pathlib import Path

import pytest
from sqlalchemy import select
from sqlalchemy.exc import NoResultFound
from sqlalchemy.orm.session import Session as SASession
from starlette.testclient import TestClient

from duty_board.alchemy.session import create_session
from duty_board.models.person import Person
from duty_board.models.person_image import PersonImage
from duty_board.server import app


@pytest.mark.usefixtures("_get_fake_dataset")
def test_get_schedule() -> None:
    client = TestClient(app)
    result = client.get("/schedule", params={"timezone": "Europe/Amsterdam"})
    result_json = result.json()
    assert result_json["config"] == {
        "timezone": "Europe/Amsterdam",
        "text_color": "white",
        "background_color": "#3C9C2D",
        "categories": ["Big Data", "Infrastructure"],
        "git_repository_url": "https://github.com/Jorricks/DutyBoard",
        "enable_admin_button": False,
        "announcement_text_color": "#FFFFFF",
        "announcement_background_color": "#FF0000",
        "announcements": [
            "This is a debug setup. Please disable this announcement in your plugin by overwriting announcements variable."
        ],
        "footer_html": '\n    Welcome to DutyBoard. <br>\n    In case you want to know more check <a href="https://github.com/Jorricks/DutyBoard">here</a>.<br>\n    Cheers!\n    ',
    }

    assert len(result_json["calendars"]) == 3
    assert result_json["calendars"][0]["uid"] == "data_platform_duty"
    assert result_json["calendars"][0]["name"] == "Data Platform Duty"
    assert (
        result_json["calendars"][0]["description"]
        == "If you have any issues with Spark, Airflow or Jupyterhub, contact these guys. They are available 24/7, however, they work in AMS hours. Only call them if there is an emergency."
    )
    assert result_json["calendars"][0]["category"] == "Big Data"
    assert result_json["calendars"][0]["order"] == 1
    assert result_json["calendars"][0]["error_msg"] == ""
    assert result_json["calendars"][0]["sync"] is False
    assert len(result_json["calendars"][0]["events"]) == 4
    assert result_json["calendars"][0]["events"][0]["start_event"] == "2023-11-28 12:46:06 CET"
    assert result_json["calendars"][0]["events"][0]["end_event"] == "2023-11-29 12:46:06 CET"
    assert result_json["calendars"][0]["events"][1]["start_event"] == "2023-11-29 12:46:06 CET"
    assert result_json["calendars"][0]["events"][1]["end_event"] == "2023-11-30 12:46:06 CET"
    assert result_json["calendars"][0]["events"][2]["start_event"] == "2023-11-30 12:46:06 CET"
    assert result_json["calendars"][0]["events"][2]["end_event"] == "2023-12-02 12:46:06 CET"
    assert result_json["calendars"][0]["events"][3]["start_event"] == "2023-12-02 12:46:06 CET"
    assert result_json["calendars"][0]["events"][3]["end_event"] == "2023-12-03 12:46:06 CET"

    assert len(result_json["persons"]) == 2
    person_values = list(result_json["persons"].values())
    assert person_values[0]["username"] == "jorrick"
    assert person_values[0]["email"] == "some_email@gmail.com"
    assert person_values[1]["username"] == "bart"
    assert person_values[1]["email"] == "bart@gmail.com"
    uids = list(result_json["persons"].keys())
    assert result_json["persons"][uids[0]]["uid"] == int(uids[0])
    assert result_json["persons"][uids[1]]["uid"] == int(uids[1])


@pytest.mark.usefixtures("_get_fake_dataset")
def test_get_person() -> None:
    client = TestClient(app)
    session: SASession
    with create_session() as session:
        person: Person = session.scalars(select(Person).where(Person.username == "bart")).one()
    result = client.get("/person", params={"person_uid": person.uid, "timezone": "Europe/Amsterdam"})
    result_json = result.json()
    assert result_json["uid"] == person.uid
    assert result_json["username"] == "bart"
    assert result_json["email"] == "bart@gmail.com"
    assert result_json["img_filename"] is None
    assert result_json["img_width"] is None
    assert result_json["img_height"] is None
    assert result_json["extra_attributes"] == [
        {
            "information": "Send Mattermost message",
            "icon": "FaExchangeAlt",
            "icon_color": "#2980B9",
            "url": "https://mattermost.com/bart",
        }
    ]
    assert result_json["last_update"] == "2023-11-28 12:46:06 CET"
    assert result_json["error_msg"] == ""
    assert result_json["sync"] is False


@pytest.mark.usefixtures("_wipe_database")
def test_empty_persons_table() -> None:
    client = TestClient(app)
    with pytest.raises(ValueError, match="Invalid person_uid=0 passed."):
        client.get("/person", params={"person_uid": 0, "timezone": "Europe/Amsterdam"})


@pytest.mark.usefixtures("_wipe_database")
def test_get_person_image() -> None:
    client = TestClient(app)
    session: SASession
    with create_session() as session:
        person_image = PersonImage(
            image_in_bytes=(Path(__file__).parent / "data" / "openldap_ldifs" / "jan.jpeg").read_bytes()
        )
        session.add(person_image)

    with pytest.raises(NoResultFound, match="No row was found when one was required"):
        client.get("/person_img/999999999")

    response = client.get(f"/person_img/{person_image.uid}")
    assert response.headers["content-type"] == "image/jpeg"
    assert response.content == person_image.image_in_bytes



from typing import TYPE_CHECKING

from sqlalchemy.orm import Mapped, mapped_column, relationship

from duty_board.alchemy.settings import Base

if TYPE_CHECKING:
    from duty_board.models.person import Person


class PersonImage(Base):
    __tablename__ = "person_image"
    uid: Mapped[int] = mapped_column(primary_key=True, autoincrement=True, unique=True)
    image_in_bytes: Mapped[bytes] = mapped_column(nullable=False)  # LargeBinary
    person: Mapped["Person"] = relationship(
        back_populates="image",
        cascade="all, delete-orphan",
        passive_deletes=True,
    )

    def __repr__(self) -> str:
        return f"PersonImage(uid='{self.uid}')"



import logging
import os
from typing import Callable

from sqlalchemy import create_engine
from sqlalchemy.engine import Engine
from sqlalchemy.orm import DeclarativeBase, scoped_session, sessionmaker
from sqlalchemy.orm import Session as SASession

log = logging.getLogger(__name__)
SQL_ALCHEMY_CONN: str = os.environ["SQL_ALCHEMY_CONNECTION"]
engine: Engine = create_engine(
    SQL_ALCHEMY_CONN,
    connect_args={},
    pool_size=10,
    pool_recycle=1800,
    pool_pre_ping=True,
    max_overflow=10,
)
Session: Callable[..., SASession] = scoped_session(
    sessionmaker(
        autocommit=False,
        autoflush=False,
        bind=engine,
        expire_on_commit=False,
    ),
)


class Base(DeclarativeBase):
    pass



import logging
import sys
from functools import wraps
from typing import (
    Callable,
    Final,
    List,
    Literal,
    Mapping,
    Optional,
    Set,
    Tuple,
    TypeVar,
)

from ldap3 import SUBTREE, Connection, Server
from ldap3.core import exceptions
from ldap3.core.exceptions import LDAPNoSuchObjectResult
from pydantic import BaseModel

if sys.version_info[:2] >= (3, 10):
    from typing import ParamSpec
else:
    from typing_extensions import ParamSpec


logger = logging.getLogger(__name__)
P = ParamSpec("P")
R = TypeVar("R")


class LdapGroupDoesNotExistError(Exception):
    pass


def ldap_rebind(func: Callable[P, R]):  # type: ignore
    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        ldap_instance = args[0]
        if not isinstance(ldap_instance, LDAPBaseClient):
            raise TypeError(f"Unexpected first argument; {args[0]}.")
        try:
            return func(*args, **kwargs)
        except (exceptions.LDAPSocketOpenError, exceptions.LDAPSessionTerminatedByServerError):
            ldap_instance.connection.unbind()
            ldap_instance.connection.bind()
            return func(*args, **kwargs)

    return wrapper


class LDAPPersonSearch(BaseModel):
    base_dn: str
    account_attribute: str
    organization_unit: Optional[str] = None
    name: str

    @property
    def search_filter(self) -> str:
        if "." and "@" in self.name:
            return f"(mail={self.name})"
        if "@" in self.name:
            return f"({self.account_attribute}={self.name.split('@')[0]})"
        return f"({self.account_attribute}={self.name})"

    @property
    def person_dn(self) -> str:
        if self.organization_unit:
            return f"ou={self.organization_unit},{self.base_dn}"
        return self.base_dn


class LDAPGroupSearch(BaseModel):
    base_dn: str
    group_attribute: str
    organization_unit: Optional[str] = None
    ldap_group_membership_relation: Tuple[str, str]
    name: str

    @property
    def search_filter(self) -> str:
        return f"(&(ObjectClass={self.ldap_group_membership_relation[0]})({self.group_attribute}={self.name}))"

    @property
    def group_dn(self) -> str:
        if self.organization_unit:
            return f"ou={self.organization_unit},{self.base_dn}"
        return self.base_dn

    @property
    def dn(self) -> str:
        return f"{self.group_attribute}={self.name},{self.group_dn}"


class LDAPBaseClient:
    def __init__(  # noqa: PLR0913
        self,
        url: str,  # e.g. ldaps://some.ldap.server.com:636
        base_dn: str,  # e.g. dc=example,dc=com
        user_organizational_unit: Optional[str],  # e.g. People
        group_organizational_unit: Optional[str],  # e.g. Group
        account_attribute: str,  # e.g. uid
        group_attribute: str,  # e.g. cn
        ldap_group_membership_relation: Tuple[str, str],  # e.g. ("groupOfUniqueNames", "uniquemember")
        full_quantified_username: str,  # e.g. uid=jorrick,ou=People,dc=example,dc=com
        password: str,
        auto_referrals: bool = False,
    ):
        self.url: Final[str] = url
        self.base_dn: Final[str] = base_dn
        self.user_organizational_unit: Final[Optional[str]] = user_organizational_unit
        self.group_organizational_unit: Final[Optional[str]] = group_organizational_unit
        self.account_attribute: Final[str] = account_attribute
        self.group_attribute: Final[str] = group_attribute
        self.ldap_group_membership_relation: Final[Tuple[str, str]] = ldap_group_membership_relation
        self.connection: Connection = self._create_connection(
            full_quantified_username,
            password,
            auto_referrals=auto_referrals,
        )

    def _create_connection(self, bind_dn: str, password: str, auto_referrals: bool) -> Connection:
        try:
            server = Server(host=self.url)
            connection = Connection(
                server=server,
                user=bind_dn,
                password=password,
                auto_referrals=auto_referrals,
                raise_exceptions=True,
            )
            connection.bind()
        except exceptions.LDAPInvalidServerError as exc:
            raise exceptions.LDAPException(f"Unable to connect to ldap url: {self.url}.") from exc
        except exceptions.LDAPInvalidCredentialsResult as exc:
            raise exceptions.LDAPException(f"Incorrect credentials for ldap url: {self.url}.") from exc
        except exceptions.LDAPInvalidDNSyntaxResult as exc:
            raise exceptions.LDAPException(f"Invalid bind user: {bind_dn}.") from exc
        if not connection.bound:
            raise exceptions.LDAPException(f"LDAP not bound for url {self.url}")
        return connection

    @ldap_rebind
    def _search(
        self,
        filter_str: str,
        base: str,
        attr_list: Optional[List[str]] = None,
        scope: Literal["BASE", "LEVEL", "SUBTREE"] = SUBTREE,
    ) -> Optional[List[Tuple[str, Mapping[str, List[str]]]]]:
        try:
            search_success = self.connection.search(base, filter_str, search_scope=scope, attributes=attr_list)
        except LDAPNoSuchObjectResult:
            return None
        if not search_success or not self.connection.response or not self.connection.response[0]:
            return None
        return [(item["dn"], item["attributes"]) for item in self.connection.response]

    def _extract_name(self, item: Optional[List[Tuple[str, Mapping[str, List[str]]]]], i: int) -> Optional[str]:
        try:
            return item[i][1][self.group_attribute][0]  # type: ignore[index]
        except (IndexError, TypeError):
            return None

    def __get_group_search(self, group_name: str) -> LDAPGroupSearch:
        return LDAPGroupSearch(
            base_dn=self.base_dn,
            group_attribute=self.group_attribute,
            organization_unit=self.group_organizational_unit,
            ldap_group_membership_relation=self.ldap_group_membership_relation,
            name=group_name,
        )

    def _get_group_cn(self, group_name: str) -> Optional[str]:
        group_search = self.__get_group_search(group_name=group_name)
        search_result = self._search(
            filter_str=group_search.search_filter,
            base=group_search.group_dn,
            attr_list=[group_search.group_attribute],
        )
        return self._extract_name(search_result, 0)

    def _get_group_members(self, group_name: str) -> Tuple[List[str], List[str]]:
        group_search = self.__get_group_search(group_name=group_name)
        search_result = self._search(
            filter_str=f"(&(ObjectClass={self.ldap_group_membership_relation[0]}))",  # e.g. "groupOfUniqueNames"
            base=group_search.dn,
            attr_list=[self.ldap_group_membership_relation[1]],
        )
        if (not search_result) and (not self.is_existing_group(group_name)):
            raise LdapGroupDoesNotExistError(f"The group {group_name} does not exist as {group_search.dn} on LDAP.")

        dn_list: List[str] = search_result[0][1][self.ldap_group_membership_relation[1]]  # e.g. "uniquemember"
        user_list: List[str] = []
        group_list: List[str] = []
        for dn in dn_list:
            if self.user_organizational_unit is not None and f"ou={self.user_organizational_unit}," in dn:
                user_list.append(dn)
            elif self.group_organizational_unit is not None and f"ou={self.group_organizational_unit}," in dn:
                group_list.append(dn)
            elif self.account_attribute != self.group_attribute:
                if dn.startswith(self.group_attribute):
                    group_list.append(dn)
                elif dn.startswith(self.account_attribute):
                    user_list.append(dn)
                else:
                    logger.warning(f"Ignoring {dn=} as it seems to be neither a group nor a user.")
            else:
                logger.warning(f"Unable to distinguish users from groups. Assuming {dn=} is a user.")
                user_list.append(dn)
        return user_list, group_list

    def get_user(
        self,
        username: str,
        attributes: Optional[List[str]] = None,
    ) -> Optional[List[Tuple[str, Mapping[str, List[str]]]]]:
        person_search: LDAPPersonSearch = LDAPPersonSearch(
            base_dn=self.base_dn,
            account_attribute=self.account_attribute,
            organization_unit=self.user_organizational_unit,
            name=username,
        )
        return self._search(  # type: ignore[no-any-return]
            person_search.search_filter,
            person_search.person_dn,
            attributes or ["*"],
        )

    def is_existing_user(self, username: str) -> bool:
        return self.get_user(username) is not None

    def is_existing_group(self, group_name: str) -> bool:
        group_cn = self._get_group_cn(group_name)
        return group_cn is not None and len(group_cn) > 0

    def get_group_users_recursively(self, group_name: str, groups_seen: Optional[Set[str]] = None) -> List[str]:
        all_users: Set[str] = set()
        groups_seen = groups_seen or set()
        users, groups = self._get_group_members(group_name)
        all_users.update([user.split(",")[0].split("=")[1] for user in users])  # uid=jorrick,dc... -> jorrick
        for group in groups:
            group_name = group.split(",")[0].split("=")[1]
            if group_name not in groups_seen:
                all_users.update(self.get_group_users_recursively(group_name, {*groups_seen, group_name}))
        return sorted(all_users)

    def is_user_in_group(self, username: str, group_name: str) -> bool:
        return (
            self.is_existing_user(username)
            and self.is_existing_group(group_name)
            and username in self.get_group_users_recursively(group_name)
        )



import datetime
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, ClassVar, List, Optional

from sqlalchemy.orm import Session as SASession

from duty_board.models.calendar import Calendar
from duty_board.models.person import Person
from duty_board.plugin.helpers.duty_calendar_config import DutyCalendarConfig


class AbstractPlugin(ABC):
    admin_session_length: ClassVar[datetime.timedelta] = datetime.timedelta(days=7)
    person_update_frequency: ClassVar[datetime.timedelta] = datetime.timedelta(hours=1)
    calendar_update_frequency: ClassVar[datetime.timedelta] = datetime.timedelta(days=1)
    background_color_hex: ClassVar[str] = "#3C9C2D"
    text_color_hex: ClassVar[str] = "white"
    absolute_path_to_favicon_ico: ClassVar[Path] = Path(__file__).resolve().parent / "example" / "favicon.ico"
    absolute_path_to_company_logo_png: ClassVar[Path] = Path(__file__).resolve().parent / "example" / "example_logo.png"
    category_order: ClassVar[List[str]] = []
    duty_calendar_configurations: ClassVar[List[DutyCalendarConfig]] = []
    enable_admin_button: ClassVar[bool] = False
    git_repository_url: ClassVar[Optional[str]] = "https://github.com/Jorricks/DutyBoard"
    footer_html: ClassVar[Optional[str]] = """
    Welcome to DutyBoard. <br>
    In case you want to know more check <a href="https://github.com/Jorricks/DutyBoard">here</a>.<br>
    Cheers!
    """

    announcement_background_color_hex: ClassVar[str] = "#FF0000"
    announcement_text_color_hex: ClassVar[str] = "#FFFFFF"
    announcements: ClassVar[List[str]] = [
        "This is a debug setup. Please disable this announcement in your plugin by overwriting announcements variable.",
    ]

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)

    @abstractmethod
    def sync_person(self, person: Person, session: SASession) -> Person:
        pass

    @abstractmethod
    def sync_calendar(self, calendar: Calendar, session: SASession) -> Calendar:
        pass

    @abstractmethod
    async def admin_login_attempt(self, username: str, password: str) -> bool:
        pass



from pendulum.datetime import DateTime
from sqlalchemy import String
from sqlalchemy.orm import Mapped, mapped_column

from duty_board.alchemy.settings import Base
from duty_board.alchemy.sqlalchemy_types import UtcDateTime


class Token(Base):
    __tablename__ = "token"
    token: Mapped[str] = mapped_column(String(50), unique=True, primary_key=True)
    username: Mapped[str] = mapped_column(String(50), unique=True)
    last_update_utc: Mapped[DateTime] = mapped_column(UtcDateTime(), nullable=False)

    def __repr__(self) -> str:
        return f"Token(token='{self.token}', last_update_utc='{self.last_update_utc}')"



import json
import pprint
import subprocess
from pathlib import Path
from typing import Any, Dict

import requests


def to_camel_case(snake_str: str) -> str:
    components = snake_str.split("_")
    # We capitalize the first letter of each component except the first one
    # with the 'title' method and join them together.
    return components[0] + "".join(x[0].upper() + x[1:] for x in components[1:])


def convert_dict_from_snake_to_camel(given_value: Any) -> Any:
    if isinstance(given_value, (bool, int, float)):
        return given_value
    if isinstance(given_value, list):
        return [
            to_camel_case(item) if isinstance(item, str) else convert_dict_from_snake_to_camel(item)
            for item in given_value
        ]
    if isinstance(given_value, dict):
        return {
            to_camel_case(key): to_camel_case(value)
            if isinstance(value, str)
            else convert_dict_from_snake_to_camel(value)
            for key, value in given_value.items()
        }
    raise ValueError(f"Unexpected value {given_value=}.")


def write_to_openapi(path_to_openapi: Path) -> None:
    with path_to_openapi.open("w") as f:
        result = requests.get("http://127.0.0.1:8000/openapi.json", timeout=5)
        openapi_definition: Dict[str, Any] = result.json()
        pprint.pprint(openapi_definition)
        openapi_camel_case = convert_dict_from_snake_to_camel(openapi_definition)
        json.dump(openapi_camel_case, f)


def generate_typescript_client() -> None:
    output_folder = "./js/api"
    output_file = "api-generated-types.ts"
    cmd = f"npx swagger-typescript-api -p openapi.json -o {output_folder} -n {output_file} --no-client"
    docker_run_cmd = f'docker exec duty_frontend /bin/bash -c "{cmd}"'
    subprocess.run(docker_run_cmd, cwd=Path.cwd().parent / "www", shell=True, check=True)  # noqa: S602


def run() -> None:
    path_to_openapi = Path.cwd().parent / "www" / "openapi.json"
    write_to_openapi(path_to_openapi)
    print(f"Written openapi.json to {path_to_openapi!s}.")
    generate_typescript_client()
    path_to_openapi.unlink()
    print("Removed openapi.json again.")


if __name__ == "__main__":
    run()



"""initial version

Revision ID: 9af1232f6bc0
Revises:
Create Date: 2023-11-28 11:13:26.408554

"""
from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op

from duty_board.alchemy.sqlalchemy_types import UtcDateTime

# revision identifiers, used by Alembic.
revision: str = "9af1232f6bc0"
down_revision: Union[str, None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "calendar",
        sa.Column("uid", sa.String(length=50), nullable=False),
        sa.Column("name", sa.String(length=200), nullable=False),
        sa.Column("description", sa.String(length=5000), nullable=True),
        sa.Column("category", sa.String(length=50), nullable=False),
        sa.Column("order", sa.Integer(), nullable=False),
        sa.Column("icalendar_url", sa.String(length=500), nullable=False),
        sa.Column("event_prefix", sa.String(length=50), nullable=True),
        sa.Column(
            "error_msg", sa.String(length=9999), nullable=True, comment="If any, the error of the latest sync attempt."
        ),
        sa.Column("last_update_utc", UtcDateTime(timezone=True), nullable=False),
        sa.Column("sync", sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint("uid"),
    )
    op.create_table(
        "person_image",
        sa.Column("uid", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("image_in_bytes", sa.LargeBinary(), nullable=False),
        sa.PrimaryKeyConstraint("uid"),
        sa.UniqueConstraint("uid"),
    )
    op.create_table(
        "token",
        sa.Column("token", sa.String(length=50), nullable=False),
        sa.Column("username", sa.String(length=50), nullable=False),
        sa.Column("last_update_utc", UtcDateTime(timezone=True), nullable=False),
        sa.PrimaryKeyConstraint("token"),
        sa.UniqueConstraint("token"),
        sa.UniqueConstraint("username"),
    )
    op.create_table(
        "person",
        sa.Column("uid", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("username", sa.String(length=50), nullable=True),
        sa.Column("email", sa.String(length=50), nullable=True),
        sa.Column("image_uid", sa.Integer(), nullable=True),
        sa.Column("img_width", sa.Integer(), nullable=True),
        sa.Column("img_height", sa.Integer(), nullable=True),
        sa.Column(
            "extra_attributes_json",
            sa.String(length=100000),
            nullable=True,
            comment="Extra attributes represented as a json.",
        ),
        sa.Column(
            "error_msg", sa.String(length=9999), nullable=True, comment="If any, the error of the latest sync attempt."
        ),
        sa.Column("last_update_utc", UtcDateTime(timezone=True), nullable=False),
        sa.Column("sync", sa.Boolean(), nullable=False),
        sa.CheckConstraint("NOT(username IS NULL AND email IS NULL)", name="user_email_not_both_none"),
        sa.ForeignKeyConstraint(["image_uid"], ["person_image.uid"], ondelete="RESTRICT"),
        sa.PrimaryKeyConstraint("uid"),
        sa.UniqueConstraint("image_uid", name="unique_image_uid"),
        sa.UniqueConstraint("uid"),
        sa.UniqueConstraint("username", "email", name="username_email_unique"),
    )
    op.create_table(
        "on_call_event",
        sa.Column("uid", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("start_event_utc", UtcDateTime(timezone=True), nullable=False),
        sa.Column("end_event_utc", UtcDateTime(timezone=True), nullable=False),
        sa.Column("calendar_uid", sa.String(length=50), nullable=False),
        sa.Column("person_uid", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(["calendar_uid"], ["calendar.uid"], ondelete="CASCADE"),
        sa.ForeignKeyConstraint(["person_uid"], ["person.uid"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("uid"),
        sa.UniqueConstraint("uid"),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("on_call_event")
    op.drop_table("person")
    op.drop_table("token")
    op.drop_table("person_image")
    op.drop_table("calendar")
    # ### end Alembic commands ###



import importlib
import importlib.machinery
import importlib.util
import logging
import os
import sys
from functools import lru_cache
from inspect import isclass
from pathlib import Path
from typing import List, Optional, Type

from duty_board.exceptions import PluginLoadingError
from duty_board.plugin.abstract_plugin import AbstractPlugin
from duty_board.plugin.example.example_plugin import ExamplePlugin

logger = logging.getLogger(__name__)


def get_plugin() -> AbstractPlugin:
    loaded_plugin = _get_plugin()
    if loaded_plugin is None:
        logger.warning("Loading the ExamplePlugin as you did not specify a Plugin. This is probably not what you want!")
        return ExamplePlugin()
    return loaded_plugin


@lru_cache
def _get_plugin() -> Optional[AbstractPlugin]:
    if "DUTY_BOARD_PLUGIN_LOCATION" not in os.environ:
        return None

    file_path = Path(os.environ["DUTY_BOARD_PLUGIN_LOCATION"]).resolve()

    if not file_path.is_file():
        raise PluginLoadingError(f"{file_path=} is not a file. Please set the plugin location correctly.")
    if file_path.suffix != ".py":
        raise PluginLoadingError(f"{file_path=} must be a Python file. Please set the plugin correctly.")
    try:
        loader = importlib.machinery.SourceFileLoader(file_path.stem, str(file_path))
        spec = importlib.util.spec_from_loader(file_path.stem, loader)
        if spec is None:
            raise PluginLoadingError(f"Unable to import {file_path.stem}.")  # noqa: TRY301
        mod = importlib.util.module_from_spec(spec)
        sys.modules[spec.name] = mod
        loader.exec_module(mod)
        logger.info("Importing plugin module %s", file_path)
        list_with_plugins: List[Type[AbstractPlugin]] = [
            m for m in mod.__dict__.values() if isclass(m) and issubclass(m, AbstractPlugin) and m != AbstractPlugin
        ]
        if len(list_with_plugins) > 1:
            if (plugin_class_name := os.environ.get("DUTY_BOARD_PLUGIN_CLASS_NAME", None)) is not None:
                logger.info(f"Originally had {list_with_plugins=} but now filtering on {plugin_class_name=}.")
                list_with_plugins = [plugin for plugin in list_with_plugins if plugin.__name__ == plugin_class_name]
                logger.info(f"Left over with {list_with_plugins=}.")
            else:
                raise PluginLoadingError(f"{file_path} contains multiple AbstractPlugin implementations.")  # noqa: TRY301
        if not any(list_with_plugins):
            raise PluginLoadingError(f"{file_path} does not contain any classes that implement the AbstractPlugin.")  # noqa: TRY301
        your_plugin = list_with_plugins[0]
        return your_plugin()
    except PluginLoadingError:
        raise
    except Exception as e:
        logger.exception("Failed to import plugin %s", file_path)
        raise PluginLoadingError(f"Unable to import plugin {file_path}") from e



import asyncio
import os
import secrets
import string
from typing import Any, ClassVar, Dict, List, Optional

from fastapi import FastAPI
from pendulum.datetime import DateTime
from sqladmin import Admin, ModelView
from sqladmin.authentication import AuthenticationBackend
from sqladmin.fields import DateTimeField
from sqlalchemy import select
from sqlalchemy.orm import Session as SASession
from starlette.requests import Request
from wtforms.fields.core import Field

from duty_board.alchemy import settings
from duty_board.alchemy.session import create_session
from duty_board.alchemy.sqlalchemy_types import utc
from duty_board.models.calendar import Calendar
from duty_board.models.on_call_event import OnCallEvent
from duty_board.models.person import Person
from duty_board.models.token import Token
from duty_board.plugin.abstract_plugin import AbstractPlugin

loop = asyncio.get_event_loop()


class AppBuilderDateTimeAwareSelector(DateTimeField):
    """This is a DateTimePicker"""

    def process_formdata(self, valuelist: List[Any]) -> None:
        """This solves the issue with editing a record. Here the timezone would show up. We don't want that!"""
        super().process_formdata(valuelist)
        if self.data is not None:  # type: ignore[has-type]
            self.data = self.data.astimezone(utc)  # type: ignore[has-type]


class PersonAdmin(ModelView, model=Person):
    column_searchable_list = [Person.username, Person.email, Person.sync]
    column_sortable_list = [Person.uid, Person.username, Person.email, Person.last_update_utc, Person.sync]
    column_list = [Person.uid, Person.username, Person.email, Person.last_update_utc, Person.sync]
    form_overrides: ClassVar[Dict[str, Field]] = {"last_update_utc": AppBuilderDateTimeAwareSelector}


class CalendarAdmin(ModelView, model=Calendar):
    column_searchable_list = [Calendar.uid, Calendar.name, Calendar.sync, Calendar.category]
    column_sortable_list = [
        Calendar.uid,
        Calendar.name,
        Calendar.last_update_utc,
        Calendar.sync,
        Calendar.category,
        Calendar.order,
        Calendar.error_msg,
    ]
    column_list = [Calendar.uid, Calendar.name, Calendar.last_update_utc, Calendar.sync]
    form_columns = [  # Adds an order to it :)
        Calendar.uid,
        Calendar.name,
        Calendar.description,
        Calendar.icalendar_url,
        Calendar.event_prefix,
        Calendar.category,
        Calendar.order,
        Calendar.error_msg,
        Calendar.last_update_utc,
        Calendar.sync,
    ]
    form_overrides: ClassVar[Dict[str, Field]] = {"last_update_utc": AppBuilderDateTimeAwareSelector}
    form_include_pk = True


class OnCallEventAdmin(ModelView, model=OnCallEvent):
    column_searchable_list = [
        OnCallEvent.calendar_uid,
        OnCallEvent.start_event_utc,
        OnCallEvent.end_event_utc,
        OnCallEvent.person_uid,
    ]
    column_sortable_list = [OnCallEvent.calendar_uid, OnCallEvent.start_event_utc, OnCallEvent.end_event_utc]
    column_list = [OnCallEvent.calendar, OnCallEvent.start_event_utc, OnCallEvent.end_event_utc, OnCallEvent.person]
    form_columns = [OnCallEvent.calendar, OnCallEvent.start_event_utc, OnCallEvent.end_event_utc, OnCallEvent.person]
    form_overrides: ClassVar[Dict[str, Field]] = {
        "start_event_utc": AppBuilderDateTimeAwareSelector,
        "end_event_utc": AppBuilderDateTimeAwareSelector,
    }
    form_include_pk = True
    form_ajax_refs = {
        "person": {
            "fields": ("username", "email"),
        },
    }


class TokenAdmin(ModelView, model=Token):
    can_create = False
    can_edit = False
    can_delete = True
    column_list = [Token.username, Token.token, Token.last_update_utc]


def add_sqladmin(app: FastAPI, plugin: AbstractPlugin) -> Admin:
    authentication_backend = MyBackend(plugin=plugin)
    admin = Admin(app=app, engine=settings.engine, authentication_backend=authentication_backend)
    admin.add_view(PersonAdmin)
    admin.add_view(CalendarAdmin)
    admin.add_view(OnCallEventAdmin)
    admin.add_view(TokenAdmin)
    return admin


class MyBackend(AuthenticationBackend):
    def __init__(self, plugin: AbstractPlugin, *_: Any, **__: Any):
        super().__init__(secret_key=os.environ["DUTY_BOARD_SECRET_KEY"])
        self.plugin = plugin

    @staticmethod
    def create_token(username: str) -> str:
        session: SASession
        with create_session() as session:
            stmt = select(Token).where(Token.username == username)
            current_token: Optional[Token] = session.scalar(stmt)
            if current_token is not None:
                current_token.token = "".join(secrets.choice(string.ascii_lowercase) for i in range(50))
                current_token.last_update_utc = DateTime.utcnow()
            else:
                current_token = Token(
                    token="".join(secrets.choice(string.ascii_lowercase) for i in range(50)),
                    username=username,
                    last_update_utc=DateTime.utcnow(),
                )
                session.merge(current_token)
        return current_token.token

    def verify_token(self, provided_token: str) -> bool:
        session: SASession
        with create_session() as session:
            stmt = select(Token).where(Token.token == provided_token)
            token: Optional[Token] = session.scalar(stmt)
            if not token:
                return False
            if token.last_update_utc < DateTime.utcnow() - self.plugin.admin_session_length:
                session.delete(token)
                return False
            return True

    async def login(self, request: Request) -> bool:
        form = await request.form()
        username = form["username"]
        password = form["password"]
        if not isinstance(username, str) or not isinstance(password, str):
            raise TypeError(
                f"Got interesting types for {username=} and {password=}, {type(username)=}, {type(password)=}",
            )

        if not await self.plugin.admin_login_attempt(username=username, password=password):
            return False

        loop = asyncio.get_event_loop()
        result = await loop.run_in_executor(None, MyBackend.create_token, username)
        request.session.update({"token": result})
        return True

    async def logout(self, request: Request) -> bool:
        request.session.clear()
        return True

    async def authenticate(self, request: Request) -> bool:
        token = request.session.get("token")
        if not await loop.run_in_executor(None, self.verify_token, token):
            request.session.clear()
            return False
        return True



import datetime
import json
from collections import defaultdict
from typing import Dict, Iterable, List, Optional, Set

from pendulum.datetime import DateTime
from pytz.tzinfo import BaseTzInfo
from sqlalchemy import select
from sqlalchemy.orm import Session as SASession

from duty_board.models.calendar import Calendar
from duty_board.models.on_call_event import OnCallEvent
from duty_board.models.person import Person
from duty_board.web_helpers.response_types import (
    PersonResponse,
    _Calendar,
    _Events,
    _ExtraInfoOnPerson,
    _PersonEssentials,
)


def format_datetime_for_timezone(dt: datetime.datetime, timezone: BaseTzInfo) -> str:
    dt_tz_aware = dt.astimezone(timezone)
    return dt_tz_aware.strftime("%Y-%m-%d %H:%M:%S %Z")


def _get_events_ending_from_now_onwards(
    session: SASession,
    all_encountered_person_uids: Set[int],
    timezone: BaseTzInfo,
) -> Dict[str, List[_Events]]:
    stmt = select(OnCallEvent).where(OnCallEvent.end_event_utc >= DateTime.utcnow())
    result: Iterable[OnCallEvent] = session.scalars(stmt).all()
    mapped: Dict[str, List[_Events]] = defaultdict(list)
    for calendar_event in result:
        mapped[calendar_event.calendar_uid].append(
            _Events(
                start_event=format_datetime_for_timezone(calendar_event.start_event_utc, timezone),
                end_event=format_datetime_for_timezone(calendar_event.end_event_utc, timezone),
                person_uid=calendar_event.person_uid,
            ),
        )
        all_encountered_person_uids.add(calendar_event.person_uid)
    return mapped


def get_calendars(session: SASession, all_encountered_person_uids: Set[int], timezone: BaseTzInfo) -> List[_Calendar]:
    events: Dict[str, List[_Events]] = _get_events_ending_from_now_onwards(
        session=session,
        all_encountered_person_uids=all_encountered_person_uids,
        timezone=timezone,
    )
    result: Iterable[Calendar] = session.scalars(select(Calendar).order_by(Calendar.order)).all()
    return [
        _Calendar(
            uid=single_calendar.uid,
            name=single_calendar.name,
            description=single_calendar.description,
            category=single_calendar.category,
            order=single_calendar.order,
            last_update=format_datetime_for_timezone(single_calendar.last_update_utc, timezone),
            error_msg=single_calendar.error_msg or "",
            sync=single_calendar.sync,
            events=events.get(single_calendar.uid) or [],
        )
        for single_calendar in result
    ]


def get_peoples_essentials(session: SASession, all_person_uids: Set[int]) -> Dict[int, _PersonEssentials]:
    result: Iterable[Person] = session.scalars(select(Person).where(Person.uid.in_(all_person_uids))).all()
    return {
        a_person.uid: _PersonEssentials(
            uid=a_person.uid,
            username=a_person.username,
            email=a_person.email,
        )
        for a_person in result
    }


def parse_extra_attributes(person_uid: int, extra_attributes_str: Optional[str]) -> List[_ExtraInfoOnPerson]:
    if extra_attributes_str is None or len(extra_attributes_str) == 0:
        return []

    extra_attributes_list = json.loads(extra_attributes_str)
    if not isinstance(extra_attributes_list, dict):
        raise TypeError(f"Expected extra_attributes field to be a dict for {person_uid=}.")

    result_list: List[_ExtraInfoOnPerson] = []
    value: Dict[str, str]
    for key, value in extra_attributes_list.items():
        if not isinstance(value, dict):
            raise TypeError(f"Expected extra_attributes[{key}] field to be a dict for {person_uid=}.")
        if "information" not in value:
            raise ValueError(f"Missing required extra_attributes[{key}][information] for {person_uid=}.")
        result_list.append(
            _ExtraInfoOnPerson(
                information=value["information"],
                icon=value.get("icon", "FaMinus"),
                icon_color=value.get("icon_color", "black"),
                url=value.get("url", None),
            ),
        )
    return result_list


def get_person(session: SASession, person_uid: int, timezone: BaseTzInfo) -> PersonResponse:
    person = session.scalars(select(Person).where(Person.uid == person_uid)).one_or_none()
    if person is None:
        raise ValueError(f"Invalid {person_uid=} passed.")

    return PersonResponse(
        uid=person.uid,
        username=person.username,
        email=person.email,
        img_filename=str(person.image_uid) if person.image_uid is not None else None,
        img_width=person.img_width,
        img_height=person.img_height,
        extra_attributes=parse_extra_attributes(person_uid, extra_attributes_str=person.extra_attributes_json),
        last_update=format_datetime_for_timezone(person.last_update_utc, timezone),
        error_msg=person.error_msg or "",
        sync=person.sync,
    )



import asyncio
import io
import json
import logging
import os
from concurrent.futures import ThreadPoolExecutor
from typing import (
    Any,
    ClassVar,
    Dict,
    List,
    Mapping,
    Optional,
    Tuple,
    Union,
)

from PIL import Image
from sqlalchemy.orm import Session as SASession

from duty_board.models.person import Person
from duty_board.models.person_image import PersonImage
from duty_board.plugin.helpers.ldap_helper import LDAPBaseClient

logger = logging.getLogger(__name__)


class LDAPPluginMixin:
    # Basic LDAP settings
    LDAP_URL: ClassVar[str] = os.environ.get("LDAP_URL", "ldap://127.0.0.1:1389")
    LDAP_BASE_DN: ClassVar[str] = "dc=DutyBoard,dc=com"
    LDAP_USER_OU: ClassVar[str] = "People"
    LDAP_GROUP_OU: ClassVar[Optional[str]] = "Groups"
    LDAP_ACCOUNT_ATTRIBUTE: ClassVar[str] = "uid"
    LDAP_GROUP_ATTRIBUTE: ClassVar[str] = "cn"
    # Group would be cn=a_group,ou=Group,dc=example,dc=com
    # With the above config you'd have:
    # - Users -> `uid=abc,ou=People,dc=example,dc=com` = f'{ACCOUNT_ATTRIBUTE}=abc,ou={LDAP_USER_OU},{LDAP_BASE_DN}'.
    # - Groups -> `cn=folks,ou=Group,dc=example,dc=com` = f'{GROUP_ATTRIBUTE}=folks,ou={LDAP_GROUP_OU},{LDAP_BASE_DN}'

    # To configure who gets access to the admin interface.
    # This can either be `"groupOfUniqueNames", "uniquemember"` or `"groupOfNames", `member`
    LDAP_GROUP_MEMBERSHIP_RELATION: ClassVar[Tuple[str, str]] = "groupOfUniqueNames", "uniquemember"
    # Allow everyone part of `cn=admin-unique-interface,ou=Group,dc=DutyBoard,dc=com` to login to the admin interface.
    LDAP_ADMIN_GROUP_NAMES: ClassVar[Tuple[str, ...]] = ("admin-unique-interface",)

    def __init__(self, *args: Any, **kwargs: Any):
        self.client: Optional[LDAPBaseClient] = None
        super().__init__(*args, **kwargs)

    def get_client(self) -> LDAPBaseClient:
        if self.client is not None:
            return self.client
        self.client = LDAPBaseClient(
            url=self.LDAP_URL,
            base_dn=self.LDAP_BASE_DN,
            user_organizational_unit=self.LDAP_USER_OU,
            group_organizational_unit=self.LDAP_GROUP_OU,
            account_attribute=self.LDAP_ACCOUNT_ATTRIBUTE,
            group_attribute=self.LDAP_GROUP_ATTRIBUTE,
            ldap_group_membership_relation=self.LDAP_GROUP_MEMBERSHIP_RELATION,
            full_quantified_username=os.environ["LDAP_FULL_QUANTIFIED_USERNAME"],
            password=os.environ["LDAP_PASSWORD"],
            auto_referrals=False,
        )
        return self.client

    @staticmethod
    def _get_jpeg_photo_from_person(
        person_attributes: Mapping[str, Union[str, List[str]]],
    ) -> Optional[Tuple[bytes, int, int]]:
        if "jpegPhoto" not in person_attributes:
            return None
        img_as_b64: bytes = person_attributes["jpegPhoto"][0]  # type: ignore
        image = Image.open(io.BytesIO(img_as_b64))
        return img_as_b64, image.width, image.height

    def _extract_user_info(self, username: str) -> Tuple[str, Mapping[str, Union[str, List[str]]]]:
        result = self.get_client().get_user(username)
        if not result:
            raise ValueError(f"Could not find user {username}.")
        if len(result) > 1:
            raise ValueError(f"Found multiple people.. {[person[0] for person in result]}.")
        person_ldap_details = result[0]
        dn, attributes = person_ldap_details
        return dn, attributes

    def _get_extra_attributes(
        self,
        username: str,  # noqa: ARG002
        attributes: Mapping[str, Union[str, List[str]]],
    ) -> Dict[str, Dict[str, str]]:
        """You can use any FontAwesome icons listed here; https://react-icons.github.io/react-icons/icons?name=fa"""
        return {
            "fullName": {
                "information": attributes["cn"][0],
                "icon": "FaUserCircle",
            },
            "location": {
                "information": attributes["l"][0],
                "icon": "FaMapMarkerAlt",
            },
        }

    def sync_person(self, person: Person, session: SASession) -> Person:  # noqa: ARG002
        filter_string: Optional[str] = person.username or person.email
        if filter_string is None:
            error_msg = f"It's not allowed to have both {person.username=} and {person.email=} be None."
            raise ValueError(error_msg)
        dn, attributes = self._extract_user_info(filter_string)
        person.username = dn.split("=")[1].split(",")[0]  # uid=abc,ou= -> extracts abc
        person.email = attributes["mail"][0]
        person.extra_attributes_json = json.dumps(self._get_extra_attributes(person.username, attributes))

        image_result = self._get_jpeg_photo_from_person(attributes)
        if image_result:
            image_in_bytes, width, height = image_result
            if person.image is None:
                person.image = PersonImage(image_in_bytes=image_in_bytes)
            else:
                person.image.image_in_bytes = image_in_bytes
            person.img_width = width
            person.img_height = height
        logger.debug(f"Updating references of {person}.")
        return person

    def _is_user_in_admin_group(self, username: str) -> bool:
        return any(self.get_client().is_user_in_group(username, group) for group in self.LDAP_ADMIN_GROUP_NAMES)

    async def admin_login_attempt(self, username: str, password: str) -> bool:  # noqa: ARG002
        loop = asyncio.get_event_loop()
        with ThreadPoolExecutor() as pool:
            result: bool = await loop.run_in_executor(pool, self._is_user_in_admin_group, username)
            return result


# Left this code here, so you can copy-paste this code to your extension and check whether your setup works.
# This requires you to set the `LDAP_FULL_QUANTIFIED_USERNAME` and `LDAP_PASSWORD` environment variables though.
if __name__ == "__main__":
    example_person = Person(username="some_username")
    plugin = LDAPPluginMixin()
    from sqlalchemy.orm import create_session

    with create_session() as session:
        plugin.sync_person(example_person, session=session)
        import pprint

        pprint.pprint(example_person)  # noqa: T203



import datetime
from typing import List

from pendulum.tz.timezone import UTC
from sqlalchemy import delete, select
from sqlalchemy.orm import Session as SASession

from duty_board.models.calendar import Calendar
from duty_board.plugin.helpers.duty_calendar_config import DutyCalendarConfig


def _create_or_update_calendar(session: SASession, calendar: DutyCalendarConfig) -> None:
    calendar_db_instance = session.scalars(select(Calendar).where(Calendar.uid == calendar.uid)).one_or_none()
    if calendar_db_instance is not None:
        calendar_db_instance.name = calendar.name
        calendar_db_instance.description = calendar.description
        calendar_db_instance.category = calendar.category  # type: ignore[assignment]
        calendar_db_instance.order = calendar.order
        calendar_db_instance.icalendar_url = calendar.icalendar_url
        calendar_db_instance.event_prefix = calendar.event_prefix
        session.merge(calendar_db_instance)
    else:
        calendar_db_instance = Calendar(
            uid=calendar.uid,
            name=calendar.name,
            description=calendar.description,
            category=calendar.category,
            order=calendar.order,
            icalendar_url=calendar.icalendar_url,
            event_prefix=calendar.event_prefix,
            error_msg=None,
            last_update_utc=datetime.datetime(1970, 1, 1, 0, 0, 0, tzinfo=UTC),
            sync=True,
        )
        session.add(calendar_db_instance)


def sync_duty_calendar_configurations_to_postgres(
    session: SASession,
    duty_calendar_configurations: List[DutyCalendarConfig],
) -> None:
    for duty_calendar_config in duty_calendar_configurations:
        _create_or_update_calendar(session, duty_calendar_config)
    all_described_calendar_uids: List[str] = [dcc.uid for dcc in duty_calendar_configurations]
    session.execute(delete(Calendar).where(Calendar.uid.not_in(all_described_calendar_uids)))



from typing import TYPE_CHECKING, List, Optional

from pendulum.datetime import DateTime
from sqlalchemy import String
from sqlalchemy.orm import Mapped, mapped_column, relationship

from duty_board.alchemy.settings import Base
from duty_board.alchemy.sqlalchemy_types import UtcDateTime

if TYPE_CHECKING:
    from duty_board.models.on_call_event import OnCallEvent


class Calendar(Base):
    __tablename__ = "calendar"
    uid: Mapped[str] = mapped_column(String(50), primary_key=True, nullable=False)
    name: Mapped[str] = mapped_column(String(200), nullable=False)
    description: Mapped[Optional[str]] = mapped_column(String(5000), nullable=True)
    category: Mapped[str] = mapped_column(String(50), nullable=False)
    order: Mapped[int]
    icalendar_url: Mapped[str] = mapped_column(String(500), nullable=False)
    event_prefix: Mapped[Optional[str]] = mapped_column(String(50), nullable=True)
    error_msg: Mapped[Optional[str]] = mapped_column(
        String(9999),
        nullable=True,
        comment="If any, the error of the latest sync attempt.",
    )
    last_update_utc: Mapped[DateTime] = mapped_column(UtcDateTime(), nullable=False)
    sync: Mapped[bool] = mapped_column(default=True)
    events: Mapped[List["OnCallEvent"]] = relationship(
        back_populates="calendar",
        cascade="all, delete-orphan",
        passive_deletes=True,
    )

    def __repr__(self) -> str:
        return f"Calendar(uid={self.uid})"



from typing import TYPE_CHECKING, List, Optional

from pendulum.datetime import DateTime
from sqlalchemy import CheckConstraint, ForeignKey, String, UniqueConstraint
from sqlalchemy.orm import Mapped, mapped_column, relationship

from duty_board.alchemy.settings import Base
from duty_board.alchemy.sqlalchemy_types import UtcDateTime
from duty_board.models.person_image import PersonImage

if TYPE_CHECKING:
    from duty_board.models.on_call_event import OnCallEvent


class Person(Base):
    __table_args__ = (
        UniqueConstraint("image_uid", name="unique_image_uid"),  # To make sure it remains a one-to-one.
        UniqueConstraint("username", "email", name="username_email_unique"),
        CheckConstraint("NOT(username IS NULL AND email IS NULL)", name="user_email_not_both_none"),
    )  # To make sure it remains a one-to-one.
    __tablename__ = "person"
    uid: Mapped[int] = mapped_column(primary_key=True, autoincrement=True, unique=True)
    username: Mapped[Optional[str]] = mapped_column(String(50), nullable=True)
    email: Mapped[Optional[str]] = mapped_column(String(50), nullable=True)

    image_uid: Mapped[Optional[int]] = mapped_column(ForeignKey("person_image.uid", ondelete="RESTRICT"), nullable=True)
    image: Mapped[Optional[PersonImage]] = relationship(
        back_populates="person",
        cascade="all, delete-orphan",
        single_parent=True,
    )

    img_width: Mapped[Optional[int]]
    img_height: Mapped[Optional[int]]
    extra_attributes_json: Mapped[Optional[str]] = mapped_column(
        String(100000),
        nullable=True,
        comment="Extra attributes represented as a json.",
    )
    error_msg: Mapped[Optional[str]] = mapped_column(
        String(9999),
        nullable=True,
        comment="If any, the error of the latest sync attempt.",
    )
    last_update_utc: Mapped[DateTime] = mapped_column(UtcDateTime(), nullable=False)
    sync: Mapped[bool] = mapped_column(default=True)

    # We don't use this field, but still good to indicate this relationship exists
    upcoming_events: Mapped[List["OnCallEvent"]] = relationship(
        back_populates="person",
        cascade="all, delete-orphan",
        passive_deletes=True,
    )

    def __repr__(self) -> str:
        return f"Person(username='{self.username}', email='{self.email}')"



import logging
import subprocess
import sys
from pathlib import Path
from typing import Any, List

import click
from alembic.config import CommandLine
from click import Context

from duty_board import worker_loop
from duty_board.alchemy import update_duty_calendars
from duty_board.alchemy.session import create_session
from duty_board.plugin.abstract_plugin import AbstractPlugin
from duty_board.plugin.helpers import plugin_fetcher

logging.basicConfig(
    stream=sys.stdout,
    level=logging.DEBUG,
    format="%(asctime)s %(name)-12s %(levelname)-8s %(message)s",
    datefmt="%m-%d %H:%M",
)
logger = logging.getLogger(__name__)


@click.group()
def cli() -> None:
    pass


@cli.command(context_settings={"ignore_unknown_options": True, "help_option_names": []})
@click.argument("args", nargs=-1, type=click.UNPROCESSED)
def db(args: List[Any]) -> None:
    path_to_alembic_ini = Path(__file__).resolve().parent / "alembic.ini"
    CommandLine(prog="DutyBoard db").main(argv=["-c", str(path_to_alembic_ini), *args])  # type: ignore[no-untyped-call]


@cli.command()
def update_calendars() -> None:
    logger.info("Updating plugin calendars in the database.")
    plugin: AbstractPlugin = plugin_fetcher.get_plugin()
    with create_session() as session:
        update_duty_calendars.sync_duty_calendar_configurations_to_postgres(
            session=session, duty_calendar_configurations=plugin.duty_calendar_configurations
        )
    logger.info("Updated the calendars you want to track.")


@cli.command()
def calendar_refresher() -> None:
    logger.info("Starting the worker to refresh the calendars.")
    plugin: AbstractPlugin = plugin_fetcher.get_plugin()
    worker_loop.enter_calendar_refresher_loop(plugin)


@cli.command()
def duty_officer_refresher() -> None:
    logger.info("Starting the worker to refresh the persons.")
    plugin: AbstractPlugin = plugin_fetcher.get_plugin()
    worker_loop.enter_duty_officer_refresher_loop(plugin)


@cli.command(name="webserver", context_settings={"ignore_unknown_options": True, "allow_extra_args": True})
@click.option("--host", default="0.0.0.0", help="The IP address range to listen for.")  # noqa: S104
@click.option("--port", default="80", type=int, help="The port to listen for.")
@click.pass_context
def webserver(ctx: Context, host: str, port: int) -> None:
    command = f"uvicorn duty_board.server:app --host {host} --port {port}"
    if ctx.args:
        command += " " + " ".join(ctx.args)
    logger.info(f"Starting webserver with {command=}")

    cwd = str(Path(__file__).parent.absolute())
    with subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=1,
        shell=True,  # noqa: S602
        text=True,
        cwd=cwd,
    ) as process:
        if process.stdout is None:
            raise ValueError("Something went wrong without opening the pipeline.")
        for line in iter(process.stdout.readline, ""):
            logger.info(line)
        return_code = process.wait()
        if return_code is not None and return_code != 0:
            raise ChildProcessError(f"Process ended with problems. {return_code=}.")


if __name__ == "__main__":
    cli()



import os
from logging.config import fileConfig

from alembic import context
from sqlalchemy import create_engine

from duty_board.alchemy.settings import Base
from duty_board.models.calendar import Calendar
from duty_board.models.on_call_event import OnCallEvent
from duty_board.models.person import Person
from duty_board.models.token import Token

# This is the Alembic Config object, which provides access to the values within the .ini file in use.
config = context.config

# Interpret the config file for Python logging. This line sets up loggers basically.
if config.config_file_name is not None:
    fileConfig(config.config_file_name)

# List all models
_ = (Calendar, Person, OnCallEvent, Token)
# add your model's MetaData object here for 'autogenerate' support
target_metadata = Base.metadata

# other values from the config, defined by the needs of env.py, can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.


def run_migrations_offline() -> None:
    """Run migrations in 'offline' mode.

    This configures the context with just a URL and not an Engine, though an Engine is acceptable here as well.  By
    skipping the Engine creation we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the script output.
    """
    url = os.environ["SQL_ALCHEMY_CONNECTION"]
    context.configure(
        url=url,
        target_metadata=target_metadata,
        literal_binds=True,
        dialect_opts={"paramstyle": "named"},
    )

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online() -> None:
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine and associate a connection with the context.
    """
    url = os.environ["SQL_ALCHEMY_CONNECTION"]
    connectable = create_engine(url)

    with connectable.connect() as connection:
        context.configure(connection=connection, target_metadata=target_metadata)
        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



from datetime import datetime, timedelta
from typing import List, Optional, Tuple

import pytest
import requests_mock
from pendulum.datetime import DateTime
from pendulum.tz.timezone import UTC
from sqlalchemy import select
from sqlalchemy.orm.session import Session as SASession

from duty_board import worker_loop
from duty_board.alchemy import update_duty_calendars
from duty_board.alchemy.session import create_session
from duty_board.models.calendar import Calendar
from duty_board.models.on_call_event import OnCallEvent
from duty_board.models.person import Person
from tests import ical_helper
from tests.conftest import get_loaded_ldap_plugin


@pytest.mark.usefixtures("_wipe_database")
def test_update_the_most_outdated_person() -> None:
    # Ingest the calendars
    with get_loaded_ldap_plugin() as example_plugin:
        session: SASession
        with create_session() as session:
            update_duty_calendars.sync_duty_calendar_configurations_to_postgres(
                session=session, duty_calendar_configurations=example_plugin.duty_calendar_configurations
            )
        # First run is when there is nothing to update yet :)
        worker_loop.update_the_most_outdated_person(example_plugin)

    # Ingest the Persons & OnCallEvents
    with create_session() as session:
        calendars: List[Calendar] = list(session.scalars(select(Calendar)).all())
        persons: List[Person] = [
            Person(
                username="jan",
                email=None,
                last_update_utc=datetime(1970, 1, 1, 0, 0, 0, tzinfo=UTC),
            ),
            Person(
                username=None,
                email="henk@tank.nl",
                last_update_utc=datetime(1970, 1, 1, 0, 0, 1, tzinfo=UTC),
            ),
            Person(
                username="henk",
                email=None,
                last_update_utc=datetime(1970, 1, 1, 0, 0, 2, tzinfo=UTC),
            ),
        ]
        session.add_all(persons)
        events: List[OnCallEvent] = [
            OnCallEvent(
                start_event_utc=DateTime.utcnow() + timedelta(days=1),
                end_event_utc=DateTime.utcnow() + timedelta(days=2),
                calendar=calendars[0],
                person=persons[0],
            ),
            OnCallEvent(
                start_event_utc=DateTime.utcnow() + timedelta(days=2),
                end_event_utc=DateTime.utcnow() + timedelta(days=3),
                calendar=calendars[0],
                person=persons[1],
            ),
            OnCallEvent(
                start_event_utc=DateTime.utcnow() + timedelta(days=3),
                end_event_utc=DateTime.utcnow() + timedelta(days=4),
                calendar=calendars[0],
                person=persons[2],
            ),
        ]
        session.add_all(events)

    # We verify the events have different persons linked to them
    with create_session() as session:
        on_call_events = list(session.scalars(select(OnCallEvent).order_by(OnCallEvent.start_event_utc)).all())
        assert on_call_events[1].person_uid != on_call_events[2].person_uid

    # We proceed to do our three runs to update all three persons
    with get_loaded_ldap_plugin() as example_plugin:
        worker_loop.update_the_most_outdated_person(example_plugin)
        worker_loop.update_the_most_outdated_person(example_plugin)
        worker_loop.update_the_most_outdated_person(example_plugin)

    # Here we verify that the users were filled in and user 1 and 2 were combined.
    with create_session() as session:
        persons = list(session.scalars(select(Person)).all())
        assert len(persons) == 2
        assert persons[0].email == "jan@schoenmaker.nl"
        assert persons[1].email == "henk@tank.nl"

        on_call_events = list(session.scalars(select(OnCallEvent).order_by(OnCallEvent.start_event_utc)).all())
        assert on_call_events[1].person_uid == on_call_events[2].person_uid


@pytest.mark.usefixtures("_wipe_database")
def test_update_the_most_outdated_calendar() -> None:
    # First run is when there is nothing to update yet :)
    with get_loaded_ldap_plugin() as example_plugin:
        worker_loop.update_the_most_outdated_calendar(example_plugin)

    # Create the calendar
    session: SASession
    with create_session() as session:
        calendar = Calendar(
            uid="data_platform_duty",
            name="Data Platform Duty",
            description="Call us for your issues with the data platform",
            icalendar_url="https://non-existing-url.com/icalendar.ics",
            category="Big Data",
            order=1,
            last_update_utc=datetime(1970, 1, 1, 0, 0, 2, tzinfo=UTC),
            sync=False,
        )
        session.add(calendar)

    # Do the actual update loop
    with get_loaded_ldap_plugin() as example_plugin, requests_mock.Mocker() as m:
        events: List[Tuple[Optional[str], Optional[str], timedelta, timedelta]] = [
            ("jan", None, timedelta(days=0), timedelta(days=1)),
            ("jan", "mailto:jan@schoenmaker.nl", timedelta(days=1), timedelta(days=1)),
            ("henkietankie", "henk@tank.nl", timedelta(days=2), timedelta(days=20)),
        ]
        ical_response = ical_helper.get_icalendar_response(events=events)
        m.get("https://non-existing-url.com/icalendar.ics", text=ical_response)

        worker_loop.update_the_most_outdated_calendar(example_plugin)

    with create_session() as session:
        on_call_events = list(session.scalars(select(OnCallEvent).order_by(OnCallEvent.start_event_utc)).all())
        assert len(on_call_events) == 3
        for on_call_event in on_call_events:
            assert on_call_event.calendar_uid == "data_platform_duty"
        assert on_call_events[0].person.username == "jan"
        assert on_call_events[0].person.email is None

        assert on_call_events[1].person.username is None
        assert on_call_events[1].person.email == "jan@schoenmaker.nl"

        assert on_call_events[2].person.username is None
        assert on_call_events[2].person.email == "henk@tank.nl"



from typing import Optional

from pydantic import BaseModel, Field
from typing_extensions import Annotated


class DutyCalendarConfig(BaseModel):
    # A unique identifier that stays the same. When renaming the name object and keeping the same uid, it won't remove
    # all your events & persons in the meantime.
    uid: Annotated[str, Field(max_length=50)]
    # The name shown on the UI.
    name: Annotated[str, Field(max_length=200)]
    # A description about who owns what.
    description: Annotated[Optional[str], Field(max_length=5000)]
    # iCalendar URL. This has to be something our request library can fetch.
    icalendar_url: Annotated[str, Field(max_length=500)]
    # The category this duty belongs in (this is what menu item it is shown for)
    category: Annotated[Optional[str], Field(max_length=50)]
    # Priority of order. The lower the number, the earlier this calendar shows up.
    order: Annotated[int, Field(default=99999, ge=0, le=9999999)]
    # Prefix before user LDAP or user email is mentioned. Example prefix; 'duty:' when calendar event; 'duty: thomas'
    event_prefix: Annotated[Optional[str], Field(max_length=50, default=None)]



from datetime import datetime
from typing import Any, Optional

import pendulum
from sqlalchemy import DateTime, TypeDecorator

utc = pendulum.tz.timezone("UTC")


class UtcDateTime(TypeDecorator):  # type: ignore[type-arg]
    """
    Almost equivalent to :class:`~sqlalchemy.types.DateTime` with
    ``timezone=True`` option, but it differs from that by:

    - Never silently take naive :class:`~datetime.datetime`, instead it
      always raise :exc:`ValueError` unless time zone aware value.
    - :class:`~datetime.datetime` value's :attr:`~datetime.datetime.tzinfo`
      is always converted to UTC.
    - Unlike SQLAlchemy's built-in :class:`~sqlalchemy.types.DateTime`,
      it never return naive :class:`~datetime.datetime`, but time zone
      aware value, even with SQLite or MySQL.
    - Always returns DateTime in UTC

    """

    impl = DateTime(timezone=True)

    cache_ok = True

    def process_bind_param(self, value: Optional[datetime], dialect: Any) -> Optional[datetime]:  # noqa: ARG002
        if value is not None:
            if not isinstance(value, datetime):
                raise TypeError("expected datetime.datetime, not " + repr(value))
            if value.tzinfo is None:
                raise ValueError("naive datetime is disallowed")
            return value.astimezone(utc)
        return None

    def process_result_value(self, value: Optional[datetime], dialect: Any) -> Optional[datetime]:  # noqa: ARG002
        """
        Processes DateTimes from the DB making sure it is always returning UTC. Not using timezone.convert_to_utc as
        that converts to configured TIMEZONE while the DB might be running with some other setting. We assume UTC
        datetimes in the database.
        """
        if value is not None:
            value = value.replace(tzinfo=utc) if value.tzinfo is None else value.astimezone(utc)
        return value



import logging
import time
import traceback
from typing import Optional, Sequence, Tuple

from pendulum.datetime import DateTime
from sqlalchemy import Select, Update, or_, select, update
from sqlalchemy.orm import Session as SASession

from duty_board.alchemy.session import create_session
from duty_board.models.calendar import Calendar
from duty_board.models.on_call_event import OnCallEvent
from duty_board.models.person import Person
from duty_board.plugin.abstract_plugin import AbstractPlugin

logger = logging.getLogger(__name__)


def get_most_outdated_person(plugin: AbstractPlugin, session: SASession) -> Optional[Person]:
    update_persons_with_last_update_before: DateTime = DateTime.utcnow() - plugin.person_update_frequency
    stmt = (
        select(Person)
        .where(Person.last_update_utc <= update_persons_with_last_update_before)
        .order_by(Person.last_update_utc)
        .limit(1)
    )
    return session.scalar(stmt)


def get_most_outdated_calendar(plugin: AbstractPlugin, session: SASession) -> Optional[Calendar]:
    update_calendars_with_last_update_before: DateTime = DateTime.utcnow() - plugin.calendar_update_frequency
    stmt: Select[Tuple[Calendar]] = (
        select(Calendar)
        .where(Calendar.last_update_utc <= update_calendars_with_last_update_before)
        .order_by(Calendar.last_update_utc)
        .limit(1)
    )
    return session.scalar(stmt)


def ensure_person_uniqueness(new_person: Person) -> Person:
    """
    This function ensures we don't have 1000 users with the except same username and or email.
    If there is a new person object with the same username and email, we wipe the old person object and replace the
    UIDS. This requires to be done in a separate session to prevent duplicate key issues.
    """
    extra_session: SASession
    with create_session() as extra_session:
        # Required to expunge, otherwise the commit will also update new_person and will raise a DuplicateKeyIndexError.
        extra_session.expunge(new_person)
        query: Select[Tuple[Person]] = select(Person).where(Person.uid != new_person.uid)
        if new_person.username and new_person.email:
            query = query.where(or_(Person.username == new_person.username, Person.email == new_person.email))
        elif new_person.username:
            query = query.where(Person.username == new_person.username)
        elif new_person.email:
            query = query.where(Person.email == new_person.email)
        else:
            raise ValueError(f"{new_person.uid=} has both {new_person.username=} and {new_person.email=} as None.")

        result: Sequence[Person] = extra_session.scalars(query).all()
        logger.debug(f"Found {len(result)=}, {result=}.")

        for same_person in result:
            logger.info(f"Updating all references of {same_person.uid=} to {new_person.uid=} in OnCallEvents table.")
            # Does this need to be executed somehow?
            stmt: Update = (
                update(OnCallEvent).where(OnCallEvent.person_uid == same_person.uid).values(person_uid=new_person.uid)
            )
            extra_session.execute(stmt)
            logger.warning(f"Deleting {same_person.uid=} {same_person=} in favor of {new_person.uid=} {new_person=}.")
            extra_session.delete(same_person)
            extra_session.commit()
    return new_person


def update_the_most_outdated_person(plugin: AbstractPlugin) -> None:
    try:
        with create_session() as session:
            person: Optional[Person]
            if (person := get_most_outdated_person(plugin=plugin, session=session)) is None:
                logger.debug("Nothing to update here :).")
                time.sleep(1)  # Avoid overload on the database.
                return

            logger.info(f"Updating {person=}.")
            try:
                person = plugin.sync_person(person=person, session=session)
                logger.debug(f"Successfully executed plugins sync_person() for {person=}.")
                person = ensure_person_uniqueness(new_person=person)
                person.error_msg = None
            except Exception:
                logger.exception(f"Failed to update {person=}.")
                person.error_msg = traceback.format_exc()
            finally:
                person.last_update_utc = DateTime.utcnow()
                session.merge(person)
        logger.info("Successfully updated the state of the person in the database.")
    except Exception:
        logger.exception("Failed to update a person in the database. There is probably some database error.")


def update_the_most_outdated_calendar(plugin: AbstractPlugin) -> None:
    try:
        with create_session() as session:
            calendar: Optional[Calendar]
            if (calendar := get_most_outdated_calendar(plugin=plugin, session=session)) is None:
                logger.debug("Nothing to update here :).")
                time.sleep(1)  # Avoid overload on the database.
                return

            logger.info(f"Updating {calendar=}.")
            try:
                calendar = plugin.sync_calendar(calendar=calendar, session=session)
                logger.debug(f"Successfully executed plugins sync_calendar() for {calendar=}.")
                calendar.error_msg = None
            except Exception:
                logger.exception(f"Failed to update {calendar=}.")
                calendar.error_msg = traceback.format_exc()
            finally:
                calendar.last_update_utc = DateTime.utcnow()
                session.merge(calendar)
        logger.info("Successfully updated the state of the calendar in the database.")
    except Exception:
        logger.exception("Failed to update some calendar in the database. There is probably some database error.")


def enter_calendar_refresher_loop(plugin: AbstractPlugin) -> None:
    while True:
        update_the_most_outdated_calendar(plugin=plugin)


def enter_duty_officer_refresher_loop(plugin: AbstractPlugin) -> None:
    while True:
        update_the_most_outdated_person(plugin=plugin)



from pathlib import Path
from typing import ClassVar, List

from duty_board.plugin.abstract_plugin import AbstractPlugin
from duty_board.plugin.helpers.duty_calendar_config import DutyCalendarConfig
from duty_board.plugin.ical_plugin_mixin import ICalPluginMixin
from duty_board.plugin.ldap_plugin_mixin import LDAPPluginMixin


class ExamplePlugin(ICalPluginMixin, LDAPPluginMixin, AbstractPlugin):
    absolute_path_to_favicon_ico = Path(__file__).resolve().parent / "favicon.ico"
    absolute_path_to_company_logo_png = Path(__file__).resolve().parent / "example_logo.png"
    category_order: ClassVar[List[str]] = ["Big Data", "Infrastructure"]
    # All required functions are implemented by ICalPluginMixin & LDAPPluginMixin

    duty_calendar_configurations: ClassVar[List[DutyCalendarConfig]] = [
        DutyCalendarConfig(
            uid="data_platform_duty",
            name="Data Platform Duty",
            description="If you have any issues with Spark, Airflow or Jupyterhub, contact these guys. "
            "They are available 24/7, however, they work in AMS hours. Only call them if there is an "
            "emergency.",
            icalendar_url="http://ical_server:8002/icalendar1.ics",
            category="Big Data",
            order=1,
            event_prefix=None,
        ),
        DutyCalendarConfig(
            uid="infrastructure_duty",
            name="Infrastructure",
            description="If you have any issues with Firewalls or critical tools such as Yarn, HDFS etc."
            "However, usually Data Platform Duty is your first point of contact for application "
            "issues.",
            icalendar_url="http://ical_server:8002/icalendar2.ics",
            category="Infrastructure",
            order=2,
            event_prefix=None,
        ),
        DutyCalendarConfig(
            uid="machine_learning",
            name="Machine learning",
            description="Do you have any issues with your machine learning tools? Ask these guys.",
            icalendar_url="http://ical_server:8002/icalendar3.ics",
            category="Big Data",
            order=3,
            event_prefix=None,
        ),
    ]
